import snowflake.connector
import pandas as pd
from dask import dataframe as dd
from typing import Dict, List, Optional, Tuple, Union
from shipyard_templates import Database, ExitCodeError, ExitCodeException
from snowflake.connector import pandas_tools as pt
from .utils import (
    _decode_rsa,
    _file_fits_in_memory,
    _get_file_size,
    _get_memory,
    map_snowflake_to_pandas,
    read_file,
)


class SnowflakeClient(Database):
    def __init__(
        self,
        username,
        pwd,
        database=None,
        account=None,
        warehouse=None,
        schema=None,
        rsa_key=None,
        role=None,
    ) -> None:
        self.username = username
        self.pwd = pwd
        self.account = account
        self.warehouse = warehouse
        self.schema = schema
        self.database = database
        self.rsa_key = rsa_key
        self.role = role
        super().__init__(
            username,
            pwd,
            account=account,
            warehouse=warehouse,
            schema=schema,
            database=database,
            rsa_key=rsa_key,
        )

    def connect(self):
        """Helper function for authentication tests to see if provided credentials are valid

        Returns: 0 for success, 1 for error

        """
        if self.rsa_key:
            if self.pwd:
                self.logger.warning(
                    "Private Key was provided in addition to Password. Using the Private Key to login"
                )
            private_key = _decode_rsa(self.rsa_key)
            try:
                con = snowflake.connector.connect(
                    user=self.username,
                    account=self.account,
                    private_key=private_key,
                    warehouse=self.warehouse,
                    database=self.database,
                    schema=self.schema,
                    role=self.role,
                )
                self.logger.info(f"Successfully connected to Snowflake")
                return con
            except Exception as e:
                self.logger.error(
                    f"Could not authenticate to Snowflake for user {self.username} with credentials in {self.rsa_key}"
                )
                self.logger.exception(e)
                return 1
        else:
            try:
                con = snowflake.connector.connect(
                    user=self.username,
                    password=self.pwd,
                    account=self.account,
                    warehouse=self.warehouse,
                    database=self.database,
                    schema=self.schema,
                    role=self.role,
                )
                self.logger.info(f"Successfully connected to snowflake")
                return con
            except Exception as e:
                self.logger.error(
                    f"Could not authenticate to account {self.account} for user {self.username}"
                )
                self.logger.exception(e)
                return 1  # failed connection

    def upload(
        self,
        conn: snowflake.connector.SnowflakeConnection,
        df: pd.DataFrame,
        table_name: str,
        if_exists: str = "replace",
    ):
        """Uploads a pandas dataframe to a snowflake table

        Args:
            conn (snowflake.connector.SnowflakeConnection): The snowflake connection object
            df (pd.DataFrame): The dataframe to be uploaded
            table_name (str): The name of the Snowflake Table to, if it doesn't exist, it will be created
            insert_method (str): The method to use when inserting the data into the table. Options are replace or append Defaults to 'replace'

        Returns:
            _type_: A tuple of the success of the upload, the number of chunks, the number of rows, and the output
        """
        if if_exists not in ["replace", "append"]:
            raise ExitCodeException(
                f"Invalid insert method: {if_exists} is not a valid insert method. Choose between 'replace' or 'append'",
                self.EXIT_CODE_INVALID_ARGUMENTS,
            )

        if if_exists == "replace":
            self.logger.info("Uploading data to Snowflake via replace")
            self.execute_query(conn, f"DROP TABLE IF EXISTS {table_name}")
            success, nchunks, nrows, output = pt.write_pandas(
                conn=conn, df=df, table_name=table_name, auto_create_table=True
            )
            self.logger.info("Successfully uploaded data to Snowflake")
            return success, nchunks, nrows, output
        else:
            self.logger.info("Uploading data to Snowflake via append")
            success, nchunks, nrows, output = pt.write_pandas(
                conn=conn, df=df, table_name=table_name
            )
            self.logger.info("Successfully uploaded data to Snowflake")
            return success, nchunks, nrows, output

    def execute_query(
        self, conn: snowflake.connector.SnowflakeConnection, query: str
    ) -> snowflake.connector.cursor.SnowflakeCursor:
        try:
            cursor = conn.cursor()
            cursor.execute(query)
            self.logger.info(f"Successfully executed query: {query} in Snowflake")
            return cursor
        except Exception as e:
            self.logger.error("Could not execute the provided query in Snowflake")
            raise ExitCodeException(e, self.EXIT_CODE_INVALID_QUERY)

    def fetch(
        self, conn: snowflake.connector.SnowflakeConnection, query: str
    ) -> pd.DataFrame:
        cursor = self.execute_query(conn, query)
        results = cursor.fetchall()
        return pd.DataFrame(results, columns=[desc[0] for desc in cursor.description])



import psutil
import os
import pandas as pd
from dask import dataframe as dd
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import dsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization.ssh import serialize_ssh_public_key
from typing import Dict, List, Optional, Tuple, Union
from copy import deepcopy


def _get_file_size(file: str) -> int:
    """Helper function to get the size of a given file

    Args:
        file: The desired file to grab the size of

    Returns: The size in bytes
    """
    return os.path.getsize(file)


def _get_memory() -> int:
    """Helper function to get the available memory

    Returns: The amount of memory in bytes

    """
    return psutil.virtual_memory().total


def _file_fits_in_memory(file: str) -> bool:
    """Helper function to show if a file can be read into memory

    Args:
        file: The desired file to be read in

    Returns: True if the file can fit in memory and False otherwise

    """
    if _get_file_size(file) < _get_memory():
        return True
    return False


def _decode_rsa(rsa_key: str):
    try:
        with open(rsa_key, "rb") as key:
            p_key = serialization.load_pem_private_key(
                key.read(),
                password=os.environ["PRIVATE_KEY_PASSPHRASE"].encode(),
                backend=default_backend(),
            )
        pkb = p_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
        return pkb

    except Exception as e:
        return None


def map_snowflake_to_pandas(snowflake_data_types: List[List]) -> Union[Dict, None]:
    """Helper function to map a snowflake data type to the associated pandas data type

    Args:
        List

    Returns:
        dict | None: Dict where the key is the field name and the value is the pandas data type
    """
    if snowflake_data_types is None:
        return None
    snowflake_to_pandas = {
        "BOOLEAN": "bool",
        "TINYINT": "int8",
        "SMALLINT": "int16",
        "INTEGER": "int32",
        "INT": "int32",
        "BIGINT": "int64",
        "FLOAT": "float32",
        "DOUBLE": "float64",
        "DECIMAL": "float64",
        "NUMERIC": "float64",
        "NUMBER": "float64",
        "REAL": "float32",
        "DATE": "datetime64[ns]",
        "TIME": "datetime64[ns]",
        "DATETIME": "datetime64[ns]",
        "TIMESTAMP": "datetime64[ns]",
        "VARCHAR": "object",
        "NVARCHAR": "object",
        "CHAR": "object",
        "NCHAR": "object",
        "BINARY": "object",
        "VARBINARY": "object",
        "STRING": "object",
    }

    pandas_dtypes = {}
    for item in snowflake_data_types:
        field = item[0]
        dtype = item[1]
        try:
            converted = snowflake_to_pandas[str(dtype).upper()]
            if converted is None:
                raise Exception(
                    f"Invalid datatypes: the datatype {field} is not a recognized snowflake datatype"
                )

            pandas_dtypes[field] = converted
        except KeyError as e:
            raise Exception(
                f"Invalid datatype: the datatype {field} is not a recognized snowflake datatype"
            )

    return pandas_dtypes


def get_pandas_dates(pandas_datatypes: dict) -> tuple:
    dates = []
    new_dict = deepcopy(pandas_datatypes)
    for k, v in pandas_datatypes.items():
        if v in ["datetime64[ns]", "datetime64"]:
            dates.append(k)
            del new_dict[k]
    return (dates, new_dict) if dates else (None, new_dict)


def read_file(
    file: str, snowflake_dtypes: Union[List, None] = None, file_type: str = "csv"
) -> pd.DataFrame:
    """Helper function to read in a file to a pandas dataframe. This will be build out in the future to allow for more file types like parquet, arrow, tsv, etc.
    Args:
        file (str): The file to be read in as a dataframe
        pandas_dtypes (Union[Dict,None]): The optional dictionary of pandas data types to be used when reading in the file. Defaults to None
        file_type (str): The type of file to be read in. Defaults to 'csv'

    Returns:
        pd.DataFrame: The dataframe output of the file
    """
    if snowflake_dtypes:
        pandas_dtypes = map_snowflake_to_pandas(snowflake_dtypes)
        dates, pandas_dtypes = get_pandas_dates(
            pandas_dtypes
        )  # get the dates to be parsed

    else:
        pandas_dtypes = None
        dates = None
    if _file_fits_in_memory(file):
        # if it fits in memory
        if file_type == "csv":  # alwys true for now, will be augmented in the future
            if dates:
                df = pd.read_csv(
                    file,
                    dtype=pandas_dtypes,
                    parse_dates=dates,
                    date_parser=lambda x: pd.to_datetime(x).tz_localize("UTC"),
                )
            else:
                df = pd.read_csv(file, dtype=pandas_dtypes)
    else:
        # if the file is larger than memory
        if dates:
            df = dd.read_csv(
                file,
                dtype=pandas_dtypes,
                parse_dates=dates,
                date_parser=lambda x: pd.to_datetime(x).tz_localize("UTC"),
            )
        else:
            df = dd.read_csv("file", dtype=pandas_dtypes)
    return df



import argparse
import shipyard_bp_utils as shipyard
import sys
from shipyard_snowflake import SnowflakeClient
from shipyard_templates import ExitCodeException


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--username", dest="username", required=True)
    parser.add_argument("--password", dest="password", required=False)
    parser.add_argument("--private-key-path", dest="private_key_path", required=False)
    parser.add_argument("--account", dest="account", required=True)
    parser.add_argument("--warehouse", dest="warehouse", required=False)
    parser.add_argument("--database", dest="database", required=True)
    parser.add_argument("--schema", dest="schema", required=False)
    parser.add_argument("--query", dest="query", required=True)
    parser.add_argument(
        "--destination-file-name",
        dest="destination_file_name",
        default="output.csv",
        required=True,
    )
    parser.add_argument(
        "--destination-folder-name",
        dest="destination_folder_name",
        default="",
        required=False,
    )
    parser.add_argument(
        "--file-header", dest="file_header", default="True", required=False
    )
    parser.add_argument("--user-role", dest="user_role", required=False, default="")
    args = parser.parse_args()
    return args


def main():
    args = get_args()
    client_args = {
        "username": args.username,
        "pwd": None if args.password == "" else args.password,
        "account": None if args.account == "" else args.account,
        "warehouse": args.warehouse,
        "schema": args.schema,
        "database": args.database,
        "rsa_key": None if args.private_key_path == "" else args.private_key_path,
        "role": None if args.user_role == "" else args.user_role,
    }

    destination_file_name = args.destination_file_name
    destination_folder_name = shipyard.files.clean_folder_name(
        args.destination_folder_name
    )
    destination_full_path = shipyard.files.combine_folder_and_file_name(
        folder_name=destination_folder_name, file_name=destination_file_name
    )
    file_header = shipyard.args.convert_to_boolean(args.file_header)
    client = SnowflakeClient(**client_args)
    try:
        conn = client.connect()
        if conn == 1:
            sys.exit(client.EXIT_CODE_INVALID_CREDENTIALS)
    except ExitCodeException as e:
        if client_args["password"] != "":
            client.logger.error(
                f"Ensure the password is correct for user {client_args['username']}"
            )
        if client_args["rsa_key"] != "":
            client.logger.error(
                f"Ensure the private key is correct for user {client_args['username']}"
            )
        sys.exit(client.EXIT_CODE_INVALID_CREDENTIALS)
    df = client.fetch(conn, args.query)
    try:
        df.to_csv(destination_full_path, index=False, header=file_header)
        client.logger.info(f"Successfuly wrote file to {destination_full_path}")
    except ExitCodeException as e:
        client.logger.error(f"Error writing file to {destination_full_path}")
        client.logger.error(e)
        sys.exit(client.EXIT_CODE_NO_RESULTS)


if __name__ == "__main__":
    main()



from shipyard_snowflake import SnowflakeClient
import unittest
import os
import pandas as pd


user = os.getenv("SNOWFLAKE_USER")
pwd = os.getenv("SNOWFLAKE_PWD")
account = os.getenv("SNOWFLAKE_ACCOUNT")
schema = os.getenv("SNOWFLAKE_SCHEMA")
database = os.getenv("SNOWFLAKE_DATABASE")
warehouse = os.getenv("SNOWFLAKE_WAREHOUSE")
rsa_key = "/Users/wespoulsen/.ssh/snowflake_key.p8"
role = os.getenv("SNOWFLAKE_ROLE")

client = SnowflakeClient(
    username=user,
    pwd=pwd,
    database=database,
    account=account,
    warehouse=warehouse,
    schema=schema,
    rsa_key=rsa_key,
)


def upload_test():
    df_path = "shipyard_snowflake/test/simple.csv"
    conn = client.connect()
    df = client.read_file(df_path)
    client.upload(conn, df=df, table_name="NEW_DATATYPES", if_exists="replace")
    print("Successfully uploaded file")


def upload_dtypes_test():
    df_path = "shipyard_snowflake/test/simple.csv"
    conn = client.connect()
    snowflake_dtypes = [
        ["string_col", "VARCHAR"],
        ["char_col", "CHAR"],
        ["int_col", "INT"],
        ["float_col", "FLOAT"],
        ["date_col", "DATE"],
        ["datetime_col", "TIMESTAMP"],
        ["bool_col", "BOOLEAN"],
    ]

    df = read_file(df_path, snowflake_dtypes=snowflake_dtypes)
    client.upload(conn, df=df, table_name="NEW_DATATYPES", if_exists="replace")


def fetch_test():
    conn = client.connect()
    df = client.fetch(conn, "SELECT * FROM NEW_DATATYPES")
    print(df.head())


def rsa_test():
    rsa_client = SnowflakeClient(
        username="BPUSER",
        pwd=None,
        rsa_key=rsa_key,
        database=database,
        account=account,
        warehouse=warehouse,
        schema=schema,
    )
    conn = rsa_client.connect()
    df = rsa_client.fetch(conn, "SELECT * FROM NEW_DATATYPES")
    print(df.head())



import argparse
import sys
import ast
import shipyard_snowflake as ss  # this is to reference the utility functions provided in the package
import snowflake.connector
import re
from shipyard_snowflake import SnowflakeClient
from shipyard_templates import ExitCodeException
from shipyard_bp_utils import files as shipyard


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--username", dest="username", required=False)
    parser.add_argument("--password", dest="password", required=False)
    parser.add_argument("--private-key-path", dest="private_key_path", required=False)
    parser.add_argument("--account", dest="account", required=True)
    parser.add_argument("--warehouse", dest="warehouse", required=False)
    parser.add_argument("--database", dest="database", required=False)
    parser.add_argument("--schema", dest="schema", default=None, required=False)
    parser.add_argument(
        "--source-file-name-match-type",
        dest="source_file_name_match_type",
        default="exact_match",
        choices={"exact_match", "regex_match"},
        required=False,
    )
    parser.add_argument(
        "--source-file-name",
        dest="source_file_name",
        default="output.csv",
        required=True,
    )
    parser.add_argument(
        "--source-folder-name", dest="source_folder_name", default="", required=False
    )
    parser.add_argument("--table-name", dest="table_name", default=None, required=True)
    parser.add_argument(
        "--insert-method",
        dest="insert_method",
        choices={"fail", "replace", "append"},
        default="append",
        required=False,
    )
    parser.add_argument(
        "--snowflake-data-types",
        dest="snowflake_data_types",
        required=False,
        default="",
    )
    parser.add_argument("--user-role", dest="user_role", required=False, default="")
    args = parser.parse_args()

    return args


def upload_multiple(
    client: SnowflakeClient,
    conn: snowflake.connector.connection.SnowflakeConnection,
    files: list,
    table_name: str,
    insert_method: str,
    pandas_dtypes: dict = None,
):
    """Uploads multiple files to a snowflake table

    Args:
        client (SnowflakeClient): The snowflake client object
        insert_method (str): The method to use when inserting the data into the table. Options are replace or append. Defaults to 'append'
        index (int, optional): The index of the file to upload. Defaults to 0.
    """
    total_files = len(files)
    for i, file_match in enumerate(files, start=1):
        client.logger.info(f"Uploading file {i} of {total_files}")
        try:
            df = ss.read_file(file_match, pandas_dtypes)

        except ExitCodeException(
            f"Error in reading file {i}", client.EXIT_CODE_FILE_NOT_FOUND
        ) as e:
            sys.exit(e.exit_code)

        try:
            client.upload(conn, df, table_name, insert_method)
        except ExitCodeException(
            f"Error in uploading file {i}", client.EXIT_CODE_INVALID_UPLOAD_VALUE
        ) as e:
            sys.exit(e.exit_code)
        if insert_method != "append":
            insert_method = "append"


def main():
    args = get_args()
    client = SnowflakeClient(
        username=args.username,
        pwd=args.password,
        account=args.account,
        rsa_key=args.private_key_path,
        warehouse=args.warehouse,
        schema=args.schema,
        database=args.database,
        role=args.user_role,
    )

    if (not client.username and not client.pwd) and (
        not client.username and not client.rsa_key
    ):
        client.logger.error(
            "Error: Either a username and password must be provided, or a username and private key file"
        )
        sys.exit(client.EXIT_CODE_INVALID_CREDENTIALS)
    conn = client.connect()  # establish connection to snowflake
    # for authtests, connection returns 1 if unsuccessful
    if conn == 1:
        sys.exit(client.EXIT_CODE_INVALID_CREDENTIALS)
    if args.snowflake_data_types:
        snowflake_data_types = ast.literal_eval(args.snowflake_data_types)
    else:
        snowflake_data_types = None

    if args.source_file_name_match_type == "regex_match":
        # match files based on pattern
        file_names = shipyard.find_all_local_file_names(args.source_folder_name)
        matching_file_names = shipyard.find_all_file_matches(
            file_names, re.compile(args.source_file_name)
        )
        client.logger.info(
            f"{len(matching_file_names)} files found. Preparing to upload..."
        )
    else:
        fp = shipyard.combine_folder_and_file_name(
            args.source_folder_name, args.source_file_name
        )
        df = ss.read_file(fp, snowflake_data_types)
        success, nchunks, nrows, output = client.upload(
            conn, df, table_name=args.table_name, if_exists=args.insert_method
        )
        client.logger.info(f"Uploaded {nrows} rows to {args.table_name}")


if __name__ == "__main__":
    main()



import argparse
import sys
from shipyard_snowflake import SnowflakeClient
from shipyard_templates import ExitCodeException


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--username", dest="username", required=True)
    parser.add_argument("--password", dest="password", required=False)
    parser.add_argument("--private-key-path", dest="private_key_path", required=False)
    parser.add_argument("--account", dest="account", required=True)
    parser.add_argument("--warehouse", dest="warehouse", required=False)
    parser.add_argument("--database", dest="database", required=True)
    parser.add_argument("--schema", dest="schema", required=False)
    parser.add_argument("--query", dest="query", required=True)
    parser.add_argument("--user-role", dest="user_role", required=False, default="")
    args = parser.parse_args()
    return args


def main():
    args = get_args()
    client_args = {
        "username": args.username,
        "pwd": None if args.password == "" else args.password,
        "account": None if args.account == "" else args.account,
        "warehouse": args.warehouse,
        "schema": args.schema,
        "database": args.database,
        "rsa_key": None if args.private_key_path == "" else args.private_key_path,
        "role": None if args.user_role == "" else args.user_role,
    }
    client = SnowflakeClient(**client_args)
    conn = client.connect()
    if conn == 1:
        sys.exit(client.EXIT_CODE_INVALID_CREDENTIALS)
    try:
        client.execute_query(conn=conn, query=args.query)
    except ExitCodeException as e:
        client.logger.error(e)
        sys.exit(client.EXIT_CODE_INVALID_QUERY)


if __name__ == "__main__":
    main()



import argparse
import os
from shipyard_blueprints import SnowflakeClient


def get_args():
    args = {}
    args["user"] = os.getenv("SNOWFLAKE_USERNAME")
    args["password"] = os.getenv("SNOWFLAKE_PASSWORD")
    args["account"] = os.getenv("SNOWFLAKE_ACCOUNT")
    return args


def main():
    args = get_args()
    user = args["user"]
    pwd = args["password"]
    account = args["account"]

    snowflake = SnowflakeClient(username=user, pwd=pwd, account=account)
    conn = snowflake.connect()
    if conn == 1:
        return 1
    else:
        return 0


if __name__ == "__main__":
    main()
